var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import brain from 'brain.js/src/index';
import { commonsTypeIsArray, commonsTypeIsObject } from 'tscommons-es-core';
import { commonsOutputAlert, commonsOutputDebug } from 'nodecommons-es-cli';
import { commonsGracefulAbortAddCallback, commonsGracefulAbortRemoveCallback } from 'nodecommons-es-process';
import { commonsFileReadJsonFileAsync, commonsFileWriteJsonFileAsync } from 'nodecommons-es-file';
import { CommonsAiTrainingLogger } from './commons-ai-training-logger';
export class CommonsAiNeuralNetwork {
    constructor() {
        this.validate = true;
        this.loggerConstructor = (instance) => new CommonsAiTrainingLogger(instance);
    }
    static validateData0To1(data) {
        if (commonsTypeIsArray(data)) {
            for (const v of data)
                if (v < 0 || v > 1)
                    throw new Error('Data is out of range. Must be 0..1');
            return;
        }
        if (commonsTypeIsObject(data)) {
            for (const k of Object.keys(data)) {
                const v = data[k];
                if (v < 0 || v > 1)
                    throw new Error('Data is out of range. Must be 0..1');
            }
            return;
        }
        throw new Error('Unknown input data type');
    }
    setValidate(validate) {
        this.validate = validate;
    }
    setLoggerConstructor(callback) {
        this.loggerConstructor = callback;
    }
    train(data, hiddenLayers, iterations = 10000, errorThresh = 0.0001, crossValidate = true) {
        return __awaiter(this, void 0, void 0, function* () {
            const logger = this.loggerConstructor(this);
            const options = {};
            commonsOutputDebug(`Using hidden layers: ${JSON.stringify(hiddenLayers)}`);
            options.hiddenLayers = hiddenLayers;
            const processedData = yield this.processIn(data);
            if (this.validate) {
                for (const d of processedData) {
                    CommonsAiNeuralNetwork.validateData0To1(d.input);
                    CommonsAiNeuralNetwork.validateData0To1(d.output);
                }
            }
            const typecast = processedData
                .map((pd) => pd);
            let trainer;
            if (crossValidate) {
                trainer = new brain.CrossValidate(brain.NeuralNetwork, options);
            }
            else {
                trainer = new brain.NeuralNetwork(options);
            }
            logger.start();
            // training can't be aborted through calls to brain.js, so the only way to stop it is to forcibly exit.
            // this isn't ideal, as other callbacks may still be running
            const callback = commonsGracefulAbortAddCallback(() => {
                commonsOutputAlert('SIGINT during brain training. This cannot be aborted. Forcing hard abort in 10 seconds');
                setTimeout(() => {
                    console.log(`10 seconds brain timeout elapsed. Forcing hard abort.`);
                    process.exit(128 + 2); // 128 + SIGINT=2
                }, 10 * 1000);
            });
            try {
                trainer.train(typecast, {
                    // tslint:disable:object-literal-sort-keys
                    log: (log) => {
                        logger.update(log);
                    },
                    logPeriod: 1,
                    iterations: iterations,
                    errorThresh: errorThresh
                    // tslint:enable:object-literal-sort-keys
                });
            }
            finally {
                commonsGracefulAbortRemoveCallback(callback);
            }
            logger.done();
            if (crossValidate) {
                this.net = trainer.toNeuralNetwork();
            }
            else {
                this.net = trainer;
            }
        });
    }
    run(data) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.net)
                throw new Error('No net loaded');
            const mapped = yield this.convertInputData(data);
            if (this.validate)
                CommonsAiNeuralNetwork.validateData0To1(mapped);
            const typecast = mapped;
            const output = this.net.run(typecast);
            return yield this.processOut(output);
        });
    }
    load(filename) {
        return __awaiter(this, void 0, void 0, function* () {
            const json = yield commonsFileReadJsonFileAsync(filename);
            if (!json)
                throw new Error('Unable to read training data');
            const net = new brain.NeuralNetwork();
            this.net = net.fromJSON(json);
        });
    }
    save(filename) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.net)
                throw new Error('No net loaded');
            yield commonsFileWriteJsonFileAsync(filename, this.net.toJSON());
        });
    }
}
//# sourceMappingURL=commons-ai-neural-network.js.map